"""
UI components designed to display progress
of processing
"""

from __future__ import annotations

from typing import Any, Optional

import enlighten
from proglog import ProgressBarLogger

_MANAGER = enlighten.get_manager()


def get_manager() -> enlighten.Manager:
    """
    Get the UI manager instance
    """
    return _MANAGER


class ProgressBar(ProgressBarLogger):
    """
    An progress bar which renders the progress of a task
    to the terminal through enlighten. It hooks into
    any function which accepts proglog progress bars
    """

    def __init__(self, description: str, indent_level: int = 0):
        super().__init__()
        self._bar: Optional[enlighten.Counter] = None
        self._desc = description
        self._indent = indent_level

    def _init_bar(self, total: int) -> None:
        self._bar = _MANAGER.counter(
            leave=False,
            total=total,
            desc=f"{self._indent * '  '}{self._desc}",
            unit="frames",
        )

    def bars_callback(
        self, _: Any, __: Any, value: int, old_value: Optional[int] = None
    ) -> None:
        if old_value is None:  # When old_value is none, this is the init value
            self._close()
            self._init_bar(value)

        elif self._bar:
            self._bar.update()

    def callback(self, **kwargs: str) -> None:
        message = kwargs.get("message", None)
        if message:
            print(f"{self._desc} - info: {message}")

    def _close(self) -> None:
        if self._bar:
            self._bar.close(clear=True)

    def __exit__(self, *_: Any) -> None:
        self._close()

    def __enter__(self) -> ProgressBar:
        return self



"""
abridge - an automated video splicing tool
"""

import ntpath
import os
from typing import List, Tuple

import numpy as np
from moviepy.editor import VideoClip, VideoFileClip

from . import ui


def _find_voids(
    clip: VideoClip, diff_threshold: float, repetition_threshold: int
) -> List[Tuple[float, float]]:
    cuts: List[Tuple[float, float]] = []
    last_used_frame = None
    last_used_frame_t = 0

    num_similar = 0

    with ui.ProgressBar(f"{clip.filename} - Processing frames", 1) as progress:
        for frame_t, frame in clip.iter_frames(with_times=True, logger=progress):
            if last_used_frame is None:
                last_used_frame = frame
                last_used_frame_t = frame_t
                continue

            diff = np.mean((frame - last_used_frame) ** 2)

            if diff < diff_threshold:
                # there are no signficant differences between this frame and the
                # last used frame

                num_similar += (
                    1  # Count the number of frames in a row which are similar
                )

            else:
                # there was a significant difference between the frames
                if num_similar > repetition_threshold:
                    # there were enough differences in a row
                    cuts.append((last_used_frame_t, frame_t))

                # take note of this frame, and use it as reference for comparison
                last_used_frame = frame
                last_used_frame_t = frame_t
                num_similar = 0

    return cuts


def _apply_cuts(clip: VideoClip, cuts: List[Tuple[float, float]]) -> VideoClip:
    for start, end in reversed(cuts):
        clip = clip.cutout(start, end)
    return clip


def abridge_clip(
    path: str,
    out_dir: str = "processed",
    diff_threshold: float = 20,
    repetition_threshold: int = 5,
) -> None:
    """
    Splice and render out the given clip with the given processing attributes
    """

    clip = VideoFileClip(path)
    clipname = ntpath.basename(path)

    cuts: List[Tuple[float, float]] = _find_voids(
        clip, diff_threshold, repetition_threshold
    )

    print(f"{clipname} - Removing empty sections")

    before_time = clip.duration

    clip = _apply_cuts(clip, cuts)

    after_time = clip.duration
    print(
        f"{clipname} - Removed {before_time - after_time}s from footage. (before: {before_time}s, after: {after_time}s)"
    )

    with ui.ProgressBar(f"{clipname} - Rendering", 1) as progress:
        clip.write_videofile(os.path.join(out_dir, clipname), logger=progress)

    clip.close()



"""
CLI module for abridge
"""

import argparse
import glob
import os
import sys
from concurrent.futures import ThreadPoolExecutor, as_completed, thread
from typing import List

from . import __name__ as module_name
from . import processor, ui


def _shutdown(executor: ThreadPoolExecutor) -> None:
    """
    Force shutdown the given threadpool
    """

    executor._threads.clear()  # type: ignore # pylint: disable=protected-access
    thread._threads_queues.clear()  # type: ignore # pylint: disable=protected-access


def _runner(args: argparse.Namespace) -> None:
    """
    Run posh splice with the given input args
    """

    clips: List[str] = []
    errors: List[str] = []

    for clip_input in args.clips:
        clip_files = glob.glob(clip_input)
        clips += clip_files

        if not clip_files:
            errors.append(f"{clip_input}: No such file or directory")

    if errors:
        sys.exit("\n".join(errors))

    if not os.path.isdir(args.out_dir):
        os.mkdir(args.out_dir)

    clips_progress = ui.get_manager().counter(
        desc="Processed clips", unit="clips", total=len(clips)
    )

    with ThreadPoolExecutor(max_workers=args.workers) as executor:
        futures = [
            executor.submit(
                processor.abridge_clip,
                clip,
                args.out_dir,
                args.diff_threshold,
                args.repetition_threshold,
            )
            for clip in clips
        ]

        try:
            for future in as_completed(futures):
                future.result()
                clips_progress.update()

        except KeyboardInterrupt:
            _shutdown(executor)

    clips_progress.close()
    ui.get_manager().stop()


def main() -> None:
    """
    CLI main, parse the input args and execute
    """

    parser = argparse.ArgumentParser(
        module_name, description="Effortlessly shorten videos"
    )
    parser.add_argument(
        "clips", metavar="clip", nargs="+", help="Clip to cut or glob group"
    )
    parser.add_argument(
        "-w",
        default=3,
        type=int,
        metavar="workers",
        help="Number of clip processors",
        dest="workers",
    )
    parser.add_argument(
        "-o", metavar="outdir", type=str, default="processed", dest="out_dir"
    )
    parser.add_argument(
        "-t",
        metavar="diff-threshold",
        type=float,
        default=20,
        dest="diff_threshold",
        help="Difference threshold required between frames for a frames to be considered different",
    )
    parser.add_argument(
        "-r",
        metavar="repetition-threshold",
        type=int,
        default=5,
        dest="repetition_threshold",
        help="Number of frames in a row required to make a cut",
    )

    _runner(parser.parse_args())



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['abridge']

package_data = \
{'': ['*']}

install_requires = \
['enlighten>=1.3,<2.0',
 'moviepy>=1.0,<2.0',
 'numpy>=1.17,<2.0',
 'proglog>=0.1.9,<0.2.0']

entry_points = \
{'console_scripts': ['abridge = abridge.cli:main']}

setup_kwargs = {
    'name': 'abridge',
    'version': '0.1.2',
    'description': 'Splice similar frames from videos',
    'long_description': '# abridge\n\n[![pipeline status](https://gitlab.com/freshollie/abridge/badges/master/pipeline.svg)](https://gitlab.com/freshollie/abridge/commits/master)\n[![coverage report](https://gitlab.com/freshollie/abridge/badges/master/coverage.svg)](http://freshollie.gitlab.io/abridge)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/python/black)\n\nEffortlessly shorten videos.\n\n## About\n\n`abridge` can automatically shorten video files by removing parts from the video\nwhere not much happens. This is great for making timelapse videos more engaging\nand removes the need for manual editing to cut these dead spots from the videos.\n\n## Installation\n\n`pip install abridge`\n\n`abridge` makes use of `moviepy`, which releys on `ffmpeg`. `ffmpeg` should be installed\nwhen the package is installed, but this may not work on some systems.\n\n## Docker\n\n`adbridge` can be run as a docker image, which gaurentees it will run\non all systems.\n\n`docker pull freshollie/abridge:latest`\n\n`docker run freshollie/abridge`\n\n## Usage\n\n```\nusage: abridge [-h] [-w workers] [-o outdir] [-t diff-threshold]\n               [-r repetition-threshold]\n               clip [clip ...]\n\nEffortlessly shorten videos\n\npositional arguments:\n  clip                  Clip to cut or glob group\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -w workers            Number of clip processors\n  -o outdir\n  -t diff-threshold     Difference threshold required between frames for a\n                        frames to be considered different\n  -r repetition-threshold\n                        Number of frames in a row required to make a cut\n```\n\n## Api\n\n```python\nfrom abridge import abridge_clip\n\nabridge_clip("/path/to/clip")\n```\n\n## Developing\n\nThe `abridge` project is managed and packaged by `poetry`\n\nUse `poetry install` to download the required packages for development\n\n`poetry run pre-commit install` should be run to install the pre-commit\nscripts which help with ensuring code is linted before push.\n\n### Tests\n\nTests are written with `pytest` and can be run with `make test`\n\n### Linting\n\n`abridge` is linted with `pylint` and formatted with `black` and `isort`\n\n`mypy` is used throughout the project to ensure consitent types.\n\n`make lint` will check linting, code formatting, and types\n\n`make format` will format code to required standards\n\n## TODO:\n\n- Test coverage on processor\n\n## License\n\n`MIT`\n',
    'author': 'Oliver Bell',
    'author_email': 'freshollie@gmail.com',
    'url': 'https://github.com/sdispater/poetry',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
